;
; huff.asm
;
; Huffman compression routines - version 1.0
; 808x assembly language, TASM compatible
; Steve Apiki, 11/90
;
; This code demonstrates algorithms discussed in the March '91 issue of BYTE.
;
; Code consists of the following blocks:
;
; 	_huff_compress	- c-language callable huffman compression routine.
;	_huff_expand 	- c-language callable huffman decompression routine
;
; compression routines:
;	add_node	- adds a node to the huffman tree
;	transmit	- transmits the code for a node
;	outbyte	- writes a literal byte to the output buffer
;	update_tree	- update the huffman tree
;	load_buffer	- load the input buffer from disk.
;	dump_buffer	- writes an output buffer to disk
;
; expansion routines:
;	add_daughter	- adds a daughter node to the huffman tree
;	incode	- reads a variable length code from the buffer
;	inbyte	- reads a literal byte
;	bit_load_buffer	- load an input buffer with variable length codes
;	dump_bit_buffer	- writes a bit oriented buffer to disk
;
; supporting routines:
;	c_swap	- node swapper for compression trees
;	e_swap	- node swapper for expansion trees
;	pop_bits	- used by transmit to pop a bitstack to a buffer
;	scale_weights	- used by update tree routines to rescale weights
; note that add_daughter calls add_node, and dump_bit_buffer calls dump_buffer.
;
DOSSEG
.MODEL COMPACT

public  _huff_compress, _huff_expand

; equates
	FILEBUFSIZ	equ	32767	; size of in and out file buffers
	NUMCHARS	equ	256	; the number of characters in the alphabet
	ROOT	equ	NUMCHARS	; special area for the root node
	MAXBITS	equ	NUMCHARS	; the maximum length of a bitstring
	NODES_IN_TREE	equ	NUMCHARS*2; twice the alphabet size
	BSTKLEN	equ	NUMCHARS/8; the length of the expansion bitstack
	RESERVE	equ	BSTKLEN+1	; reserved bytes for variable input
	TNODELEN	equ	4	; the number of bytes in a tree node
	WLISTLEN	equ	4	; the number of bytes in a weightlist entry
	HCOBUFLEN	equ	FILEBUFSIZ-((MAXBITS/8)+1) ;  the logical length of the compression output buffer
	HEOBUFLEN	equ	FILEBUFSIZ	; the logical length of the expansion buffer
	MSB_ONLY	equ	1000000000000000b	; mask
	NOT_MSB	equ	0111111111111111b	; mask
	THREE_LSBS	equ	0000000000000111b	; mask

.DATA
; variables in common to these routines

	output_seg	dw	? 	; segment of output buffer
	input_seg	dw	?	; segment of input buffer

; the tree segment, which is dynamically allocated, consists of 512 dwords,
; addressed like this:
; msb of low word	= bit assigned to this node
; lower nine bits of low word	= pointer to weight value in the weight list segment
; high word	= parent pointer in the tree
;
; the (dynamically allocated) weight list segment is addressed like this
; high word 	= weight assigned to this node
; low word 	= pointer back to tree node

	tree_seg	dw	?	; segment for huffman compression tree
	weight_list_seg	dw	?	; segment for weight list

	next_weight	dw	0	; next available spot in the weightlist

; huff-expand shares trees and weight lists with huff compress.
; however, huff-expand requires an additional structure which contains
; daughter pointers, since this routine must navigate down the tree.
; these are: two word nodes - low word is zero child, high word is one child

	child_ptr_seg	dw	?	; segment for holding daughter pointers

; finally, these are global to avoid a lot of overhead in output blocks

	totalbytes	dw	0	; bytes we've written, total
	dw	0	;it's a double word
	out_bitshift	db	0	; the current bit offset
	in_bitshfit	db	0	; the current bit offset for inputs

;-------------------------------------------------------
; huff_compress
; long huff_compress(int infh, int outfh);
; c-callable huffman compressor
; returns size of output file (bytes)
; returns zero on error
; 
; infh -- file handle of input file (must be opened for binary reads)
; outfh -- file handle of output file (must be opened for binary writes)
;-------------------------------------------------------

; huff_compress data
	hc_ibuf_end	dw	?

	hc_ifh	dw	?	; input file handle
	hc_ofh	dw	?	; output file handle

	el_ptr	dw	ROOT	; where is the empty leaf
	num_labels	dw	1	; how many leaves in the tree

.CODE
_huff_compress	proc	near
;
	dummy0	struc	; to address passed in parameters
	bptr0	dw	?
	raddr0	dw	?
	infh	dw	?
	outfh	dw	?
	dummy0	ends
huffc	equ	[bp-bptr0]

	push 	bp	; stack frame manipulations
	mov 	bp,sp
	sub	sp,bptr0

	push	si
	push	di	; preserve index registers for turbo c.

; initialize local variables

	mov	el_ptr, ROOT
	mov	num_labels, 1
	mov	next_weight, 0
	mov	out_bitshift, 0

; clear return value

	xor	ax, ax
	mov	totalbytes, ax
	mov	totalbytes+2, ax

; allocate memory for file buffers and tables 

	xor	cx, cx	; keep track of what we alloc

	mov	bx, (FILEBUFSIZ/16)+1	; get buffer for each file
	mov	ah, 48h
	int	21h
	jc	HCMemError
	inc	cl
	mov	input_seg, ax	; input buffer

	mov	ah, 48h
	int	21h
	jc	HCMemError
	inc	cl
	mov	output_seg, ax	; output buffer

	mov	ah, 48h
	mov	bx, ((NODES_IN_TREE*TNODELEN)/16)+1

	int	21h
	jc	HCMemError
	inc	cl
	mov	tree_seg, ax	; huffman compression tree

	mov	ah, 48h
	int	21h
	jc	HCMemError
	mov	weight_list_seg, ax	; weight list

	jmp	short	AllocDone

HCMemError:
	cmp	cx, 0	; how many blocks to return?
	jne	HCMEr0
	jmp	HCErrorExit0

HCMEr0:
	cmp	cx, 1
	jne	HCMEr1
	jmp	HCErrorExit1

HCMEr1:
	cmp	cx, 2
	jne	HCMEr2
	jmp	HCErrorExit2

HCMer2:
	jmp	HCErrorExit3

AllocDone:
; set up file handles

	mov	ax, huffc.infh	; save local copies of file handles
	mov	hc_ifh, ax
	mov	ax, huffc.outfh
	mov	hc_ofh, ax

; initialize the tree. mark all character locations as unused

	mov	es, tree_seg
	mov	cx, (NUMCHARS*TNODELEN)/2	; mark the character locations, the
	xor	ax, ax	; first NUMCHARS entries. although we overwrite the
	mov	di, ax	; other part of the structure, it doesn't matter
	rep	stosw

; prepare file buffers

	xor	si, si	; input buffer offset
	xor	di, di	; output buffer offset

; prepend the file with the file length so the decompressor knows when to stop.
; determine the length of the input file

	mov	ax, 4202h	; dos call - move pointer
	mov	bx, hc_ifh
	xor	cx, cx
	xor	dx, dx

	int	21h	; make call

; now hiword of length is in dx, loword in ax. write to file

	mov	es, output_seg
	stosw
	mov	ax, dx
	stosw

; now put pointer back

	mov	ax, 4200h	; dos call - move pointer
	mov	bx, hc_ifh
	xor	cx, cx
	xor	dx, dx

	int	21h	; make call

; load input buffer

	mov	dx, input_seg	; load the input buffer
	mov	bx, hc_ifh	; input handle
	call	load_buffer
	mov	hc_ibuf_end, dx	; the size of the input buffer


; ready to enter main compression loop
CompLoop:

	mov	es, input_seg
	mov	bl, es:[si]	; bl <- char to compress
	xor	bh, bh

	inc	si	; point to next character

	cmp	si, hc_ibuf_end	; check for the end of the buffer
	jbe	Cont0	; if not, continue reading from inbuffer

; if it is, reload the buffer and try again

	push	bx	; save parameters
	push	dx

	mov	bx, hc_ifh
	mov	dx, input_seg	; load the input buffer
	call	load_buffer	; try to reload
	mov	hc_ibuf_end, dx	; load_buffer returns the end of the new buffer. save it.

	pop	dx
	pop	bx

	jc	JumpOff	; done with main loop
	jmp	short	CompLoop	; ok, we're reloaded. read in a real character

JumpOff:
	jmp	EOLoop

Cont0:

; do we know this character already - look at the parent pointer of this
; character

	mov	es, tree_seg
	mov	ax, bx	; save character for later use
	shl	bx, 1
	shl	bx, 1	; bx+2 is this nodes parent ptr
	cmp	word ptr es:[bx+2], 0	; if zero, unknown
	jz	NotKnown

; this character is already known to compressor and decompressor. transmit its code

	mov	bx, outfh	; output file handle
	mov	dx, HCOBUFLEN	; length of output buffer
	mov	es, output_seg
	call	transmit	; send it
	mov	cx, offset c_swap	; hand swapping function to updater
	call	update_tree	; update the tree with new statistics
	jmp	short	CompLoop	; get another char and do it again

NotKnown:
; this character is new. update the tree by adding a new node.

	push	ax
	mov	ax, el_ptr
	mov	bx, outfh	; output file handle
	mov	dx, HCOBUFLEN	; length of output buffer
	mov	es, output_seg
	cmp	ax, ROOT	; special case -- don't transmit the empty leaf
	je	FirstCode	; if it's the root
	call	transmit	; transmit the empty leaf code
	mov	cx, offset c_swap	; hand swap function to updater
	call	update_tree	; update with current statistics

FirstCode:
	pop	ax
	call	outbyte	; transmit the full character

	cmp	num_labels, NUMCHARS	; are all characters known?
	je	LastCode	; then assign empty leaf to it

; this is not the last code.
; add the new code as child of the old one, on branch one
; add a new empty leaf as child of the old one, on branch zero

	inc	num_labels	; about to add another leaf

	mov	dx, el_ptr	; parent of this node
	mov	bx, ax	; ax <- character to add
	mov	al, 1	; branch one
	call	add_node	; add this character

	xor	al, al	; branch zero
	inc	el_ptr	; this is the new empty leaf
	mov	bx, el_ptr	; prepare to add it
	call	add_node	; add it

	jmp	CompLoop	; get another char and do it again

; last character. don't need the empty leaf any more. assign the last char to
; the empty leaf
LastCode:

	mov	es, tree_seg
	mov	bx, el_ptr
	shl	bx, 1
	shl	bx, 1	; bx+2 points to the empty leaf's parent
	mov	dx, es:[bx+2]	; dx <- emptyleaf.parent_ptr
	mov	cx, es:[bx]	; cx <- emptyleaf.weight_ptr
	mov	bx, ax	; bx <- location to add
	shl	bx, 1
	shl	bx, 1
	mov	es:[bx], cx	; save weight ptr
	mov	es:[bx+2], dx	; save parent ptr

	mov	es, weight_list_seg	; update the weight list
	mov	bx, next_weight	; bx points to empty leaf +1
	shl	bx, 1
	shl	bx, 1
	mov	es:[bx-2], ax	; change to this leaf

	jmp	CompLoop	; loop again

EOLoop:
; almost done.
; round off or don't round off the last byte in the buffer

	cmp	out_bitshift, 0	; if we finished on a boundary, don't round
	jz	NoRounding
	inc	di	; make sure we include the partial byte

NoRounding:
; dump the output buffer

	mov	bx, hc_ofh
	mov	es, output_seg
	call	dump_buffer	; write the buffer to disk

; return allocated memory
	mov	ah, 49h	; return weightlist
	mov	es, weight_list_seg
	int	21h

HCErrorExit3:
	mov	ah, 49h	; return tree seg
	mov	es, tree_seg
	int	21h

HCErrorExit2:
	mov	ah, 49h	; return output buffer
	mov	es, output_seg
	int	21h

HCErrorExit1:
	mov	ah, 49h	; return input buffer
	mov	es, input_seg
	int	21h

HCErrorExit0:
	mov	ax, totalbytes	; return value
	mov	dx, totalbytes+2	; high order word

	pop	di	; restore the registers.
	pop	si
	pop	bp

    	ret

_huff_compress	endp

;-------------------------------------------------------
; huff_expand
; long huff_expand( int infh, int outfh);
; c-callable huffman decompressor
; returns size of output file (bytes)
; returns zero on error
; 
; infh -- file handle of input file (must be opened for binary reads)
; outfh -- file handle of output file (must be opened for binary writes)
;-------------------------------------------------------
; huff_expand data
.DATA
	he_ibuf_end	dw	?

	he_ifh	dw	?	; input file handle
	he_ofh	dw	?	; output file handle

	e_el_ptr	dw	ROOT	; where is the empty leaf

	in_bitshift	db	?	; variable input bits

	target_len	dw	?	; required to find end of file
	dw	?	;it's a double word

	bytes_writn	dw	0	; bytes we've written
	dw	0	;it's a double word
;	(this is different from totalbytes to avoid undue overhead in the compressor)

.CODE
_huff_expand	proc	near
;
	dummy1	struc	; to access variable on the stack
	bptr1	dw	?
	raddr1	dw	?
	infh	dw	?
	outfh	dw	?
	dummy1	ends

huffe	equ	[bp-bptr1]

	push 	bp	; stack frame manipulations
	mov 	bp,sp
	sub	sp,bptr1

	push	si
	push	di	; preserve index registers for turbo c.

; initialize local variables

	mov	e_el_ptr, ROOT
 	xor	ax, ax
	mov	in_bitshift, al
	mov	next_weight, ax
	mov	totalbytes, ax
	mov	totalbytes+2, ax
	mov	bytes_writn, ax
	mov	bytes_writn+2, ax

; allocate memory for file buffers and tables 

	xor	cx, cx	; keep track of what's allocated

	mov	bx, (FILEBUFSIZ/16)+1	; get buffer for each file
	mov	ah, 48h
	int	21h
	jc	HEMemError
	inc	cl
	mov	input_seg, ax	; input buffer

	mov	ah, 48h
	int	21h
	jc	HEMemError
	inc	cl
	mov	output_seg, ax	; output buffer

	mov	ah, 48h
	mov	bx, ((NODES_IN_TREE*TNODELEN)/16)+1

	int	21h
	jc	HEMemError
	inc	cl
	mov	tree_seg, ax	; huffman expansion tree

	mov	ah, 48h
	int	21h
	jc	HEMemError
	inc	cl
	mov	weight_list_seg, ax	; weight list

	mov	ah, 48h
	int	21h
	jc	HEMemError
	mov	child_ptr_seg, ax	; list of daughter pointers

	jmp	short	E_AllocDone

HEMemError:
	cmp	cx, 0	; how many blocks to return?
	jne	HEMEr0
	jmp	HEErrorExit0

HEMEr0:
	cmp	cx, 1
	jne	HEMEr1
	jmp	HEErrorExit1

HEMEr1:
	cmp	cx, 2
	jne	HEMEr2
	jmp	HEErrorExit2

HEMEr2:
	cmp	cx, 3
	jne	HEMEr3
	jmp	HEErrorExit3

HEMEr3:
	jmp	HEErrorExit4

E_AllocDone:
; set up file handles

	mov	ax, huffe.infh	; save local copies of file handles
	mov	he_ifh, ax
	mov	ax, huffe.outfh
	mov	he_ofh, ax

; initialize the tree. mark all character locations as unused

	mov	es, tree_seg
	mov	cx, (NUMCHARS*TNODELEN)/2	; mark the character locations, the
	xor	ax, ax	; first NUMCHARS entries. if we overwrite the
	mov	di, ax	; other part of the structure, it doesn't matter
	rep	stosw

; prepare file buffers

	xor	si, si	; input buffer offset
	xor	di, di	; output buffer offset

	mov	dx, input_seg	; load the input buffer
	mov	bx, he_ifh	; input handle
	call	bit_load_buffer

	mov	he_ibuf_end, dx	; the size of the input buffer

; read in the target length

	mov	es, input_seg	; aim at the input buffer
	mov	ax, es:[si]	; low word of target length
	inc	si
	inc	si
	mov	dx, es:[si]	; high word of target length
	inc	si
	inc	si	; point to next word in the buffer

	mov	target_len, ax
	mov	target_len+2, dx	; store the target lengths

; read in the first character

	mov	al, es:[si]	; read the first character
	inc	si

; write out this code

	mov	es, output_seg
	stosb
	inc	bytes_writn	; keep track of bytes we've written

; add two daughters to the root -- one for the empty leaf, one for the
; new code

	mov	dx, ROOT	; the parent
	xor	bh, bh
	mov	bl, al	; which location to add
	mov	al, 1	; which branch

	call	add_daughter	; add it

	inc	e_el_ptr
	mov	bx, e_el_ptr	; new location for the leaf
	xor	al, al	; zero branch

	call	add_daughter	; add the empty leaf

; now for the main loop

ExpandLoop:

; determine if we've reached the end

	mov	ax, bytes_writn
	cmp	ax, target_len	; low orders match?
	jne	NotDone	; no
	mov	ax, bytes_writn+2
	cmp	ax, target_len+2	; high orders match?
	jne	NotDone	; no

; done all we need to do. clean up

	jmp	E_EOLoop

NotDone:
; read in a code

	mov	es, input_seg
	call 	incode	; ax <- input code

; check for buffer overflow

	cmp	si, he_ibuf_end	; is this the end of the buffer
	jbe	E_Cont	; if not, continue reading from inbuffer

; if it is, reload the buffer -- be sensitive to bit offsets

	mov	dx, input_seg	; where to load from
	mov	bx, he_ifh	; input handle
	call	bit_load_buffer	; try to reload
	mov	he_ibuf_end, dx	; the size of the input buffer

	jnc	E_Cont	; okay, reloaded
	jmp	E_EOLoop	; done - out of buffer

E_Cont:
; continue to process input code, which is in ax

	or	ah, ah	; if > 255,it's the empty leaf
	jz	CharCode

;it's the empty leaf. read in another character

	mov	cx, ax	; cx <- empty leaf pointer
	call	inbyte	; al <- character

	mov	es, output_seg
	stosb	; write out the character
	inc	bytes_writn
	jnz	NoCary
	inc	bytes_writn+2	; update bytes wrote
NoCary:

; is this the last leaf?

	cmp	e_el_ptr, NODES_IN_TREE-1	; is the empty leaf the last leaf?
	je	E_LastLeaf

; add two more nodes, one for the new empty leaf, one for the new char

	mov	dx, cx	; the parent
	xor	bh, bh
	mov	bl, al	; which location to add
	mov	al, 1	; which branch

	call	add_daughter	; add it

	inc	e_el_ptr
	mov	bx, e_el_ptr	; new location for the leaf
	xor	al, al	; zero branch

	call	add_daughter	; add the empty leaf

	mov	ax, cx	; ax <- old empty leaf ptr
	jmp	short	NonCharCode	; go on to update

E_LastLeaf:
; turn the empty leaf into the new leaf

	mov	es, tree_seg
	mov	bx, ((NODES_IN_TREE-1)*TNODELEN)	; empty leaf location
	mov	dx, es:[bx]	; dx <- empty leaf's bit/ weight ptr
	mov	cx, es:[bx+2]	; cx <- empty leaf's parent ptr

	xor	ah, ah
	mov	bx, ax
	shl	bx, 1
	shl	bx, 1	; bx <- pointer to location to add

	mov	es:[bx], dx	; save bit, weight ptr
	mov	es:[bx+2], cx	; save parent ptr

; update daughter ptrs

	mov	es, child_ptr_seg	; es:[bx] is daughter pointer
	mov	word ptr es:[bx], -1
	mov	word ptr es:[bx+2], -1	; mark this as a leaf

; update parent's daughter pointer

	mov	bx, cx	; parent pointer
	shl	bx, 1
	shl	bx, 1
	mov	es:[bx], ax	; update it

; and the weight list

	mov	es, weight_list_seg	; update the weight list
	mov	bx, next_weight	; bx points to empty leaf+1
	shl	bx, 1
	shl	bx, 1
	mov	es:[bx-2], ax	; change to this leaf

	jmp	short	NonCharCode

CharCode:
	mov	es, output_seg	; write out the code
	stosb
	inc	bytes_writn
	jnz	NoCary2
	inc	bytes_writn+2	; update bytes wrote
NoCary2:

NonCharCode:
; check for output buffer overflow

	cmp	di, HEOBUFLEN
	jb	HE_BufferOK

; dump the buffer, if full

	mov	bx, he_ofh
	call	dump_buffer

HE_BufferOK:

; now update tree to stay in sync with compressor. location to update
; is still in ax

	mov	cx, offset e_swap	; hand swapping function to update proc
	call	update_tree	; update it
	jmp	ExpandLoop

E_EOLoop:
; read all the input codes -- dump the buffer and clean up

; dump the output buffer

	mov	bx, he_ofh
	mov	es, output_seg
	call	dump_buffer

; return allocated memory

	mov	ah, 49h	; return pointer list
	mov	es, child_ptr_seg
	int	21h

HEErrorExit4:
	mov	ah, 49h	; return weightlist
	mov	es, weight_list_seg
	int	21h

HEErrorExit3:
	mov	ah, 49h	; return tree seg
	mov	es, tree_seg
	int	21h

HEErrorExit2:
	mov	ah, 49h	; return output buffer
	mov	es, output_seg
	int	21h

HEErrorExit1:
	mov	ah, 49h	; return input buffer
	mov	es, input_seg
	int	21h

HEErrorExit0:
	mov	ax, totalbytes	; return value
	mov	dx, totalbytes+2	; high order word

	pop	di	; restore the registers.
	pop	si
     pop	bp

    	ret

_huff_expand	endp

;---------------------------------------------------------------------------
; add_node
; add a node to the huffman tree. each node consists of a bit, a 
; weight list pointer, and a parent pointer. in addition, the physical
; location of the node represents its value.
; this routine adds a node, and makes the weight list pointer point to
; a valid weight of one or zero at the bottom of the current list.
; on entry:
;	al - bit
;	bx - location of the new node, its value
;	dx - parent pointer for this node
;	global next_weight holds pointer to next empty slot in weight list
; on exit
;	node added to tree
;	weight list and next_weight updated
;---------------------------------------------------------------------------
add_node	proc	near

	push	es	; preserve these registers
	push	bx

	push	bx	; save the tree location

; point es:bx to weight_list/ bit assignment word
	mov	es, tree_seg	; es points to compression tree
	shl	bx, 1
	shl	bx, 1	; double word table

; make the weight_list/ bit assignment word in ax

	ror	ax, 1	; mov lsb to msb
	and	ax, MSB_ONLY	; mask out everything else
	or	ax, next_weight	; lower nine bits = weight pointer
	inc	next_weight	; update weight pointer

	mov	es:[bx], ax	; save the word
	mov	es:[bx+2], dx	; and the parent pointer

; tree updated. now update weight list

	mov	es, weight_list_seg
	mov	bx, ax	; es:bx<- pointer to weight list
	shl	bx, 1
	shl	bx, 1	; double word table

; update low-order word, which is weight

	pop	ax	; recover the tree location

; if internal node (empty leaf), weight zero -- if label, weight one

	mov	word ptr es:[bx], 0	; set the weight to zero
	or	ah, ah	; if >255,it's an internal node
	jnz	InternalNode
	inc	word ptr es:[bx]	; if not internal, make weight one

InternalNode:

; update high order word in the weight list, which points back to tree

	mov	es:[bx+2], ax	; save tree pointer

; done adding node

	pop	bx
	pop	es	; recover registers

	ret
add_node	endp

;---------------------------------------------------------------------------
; transmit
; transmit takes a value (which corresponds to a tree location) and sends its
; code to the output buffer. To do this, it starts at the location and walks up
; the huffman tree, pushing bits on a stack. when it reaches the root, it pops
; the stack to the output buffer.
; on entry:
; 	ax - char to transmit
; 	bx - file handle
; 	dx - length of output buffer
; 	es:di - points to output buffer location
; 	bitshift is in out_bitshift global
; on exit:
;	es:di - points to new output buffer location
;	bitshift updated
;	all other registers preserved.
;---------------------------------------------------------------------------
.DATA
	bitstack	db	BSTKLEN dup (?)	; stack for output bits
	db	0	; a clear byte at the end 

.CODE
transmit	proc	near

	push	ax	; save the registers
	push	si

	push	bx	; these will be needed when popping
	push	dx	; bits to the output buffer
	push	es
	push	di

	mov	di, offset bitstack+(BSTKLEN-1)	; ds:di points to the bit stack top (end)

	xor	dx, dx	; temporary stack location
	xor	cx, cx	; keep track of number of bits in cx
	mov	es, tree_seg

TransLoop:
	push	ax	; location to transmit - save for later
	mov	bx, ax	; bl <- location to transmit
	shl	bx, 1
	shl	bx, 1	; bx points to weight pointer/ bit word
	mov	ax, es:[bx]	; ax <- weight pointer/ bit
	and	ax, MSB_ONLY	; knock off weight pointer - bit is now msb of ah
	rol	ax, 1	; bit is lsb of ax

; push this bit onto the bit stack

	shl	dl, 1	; shift the current stack byte
	or	dl, al	; dl <- current stack byte
	inc	cx
	test	cl, THREE_LSBS	; is this a multiple of eight?
	jnz	XM_NoDump	; if it is, send dl out to stack

	mov	[di], dl	; send out to stack
	xor	dl, dl	; and clear dl
	dec	di	; and update stack pointer

XM_NoDump:
	pop	bx	; recover tree location
	shl	bx, 1
	shl	bx, 1

	mov	ax, es:[bx+2]	; ax <- parent of this node
	cmp	ax, ROOT	; if it's the root, we're done
	je	DoneTransLoop
	jmp	TransLoop	; walk up another node

DoneTransLoop:
; bit stack has been completely built
; pop the bit stack
	test	cl, THREE_LSBS	; is the offset a multiple of eight
	jnz	X_HasOffset
	inc	di	; point di back at the full byte
	jmp	short	X_DoneOffsets

X_HasOffset:
	mov	[di], dl	; send out the last byte to the bitstack

X_DoneOffsets:

	mov	si, di	; si <- bitstack location
	pop	di
	pop	es	; es:di points to output buffer
	pop	dx	; dx <- output length
	pop	bx	; bx <- file handle

	call	pop_bits	; pop the bit stack to the output buffer

	pop	si	; recover registers
	pop	ax
	ret

transmit	endp

;---------------------------------------------------------------------------
; outbyte
; write out one byte (8 bits) to a bit-oriented file buffer
; on entry:
; 	al - char to write
; 	bx - file handle
; 	dx - length of output buffer
; 	es:di points to output buffer location
; 	bitshift is in out_bitshift global
; on exit:
;	es:di points to new output location.
;---------------------------------------------------------------------------
outbyte	proc	near

	push	ax
	push	bx
	push	cx

; if we're getting near the buffer limit, dump
; the buffer while we have the chance

	cmp	di, dx	; see if this is the end of the buffer
	jbe	B_DontDump
	push	ax
	mov	ax, dx
	call	dump_bit_buffer	; dump the buffer if necessary
	pop	ax

B_DontDump:
; if current shift is not zero, need to shift 
	mov	cl, out_bitshift
	or	cl, cl
	jnz	B_HasOffset

	stosb	; store the byte and we're done
	jmp	short	OByteExit

B_HasOffset:
	mov	bl, es:[di]	; bl has the current byte from the buffer

	xor	ah, ah
	shl	ax, cl	; ax has the shifted code
	or	al, bl	; al has the first byte to write out
	stosb	; write it out
	mov	es:[di], ah	; write out the high byte, too

OByteExit:
	pop	cx	; restore the registers
	pop	bx
	pop	ax
	ret

outbyte	endp

;---------------------------------------------------------------------------
; update_tree 
; Update the huffman compression tree. Start at the node passed in. Increment
; its weight. If this weight approaches the overall weight limit (an int) scale
; all weights.
; If it has become heavier than a node listed higher in the weight list, swap
; this node with the highest node that is lighter than this node.
; Exception: if the designated swap node is the parent of this node, don't swap.
; In swapping, we exchange weights and parents only, keeping all child structures
; intact.
; Continue walking up the tree, incrementing and swapping, until we reach the
; root.
;
; This dynamic huffman update algorithm is based on a technique described in
; Storer, _Data_Compression:_Methods_and_Theory_ (Computer Science Press, 1988)
; and introduced elsewhere.
;
; on entry:
;	ax - start location in tree.
;	cx - pointer to swapping function
; on exit:
;	tree updated, all registers preserved.
;---------------------------------------------------------------------------
.DATA
	swapping_function	dw	?	; address of swap routine

.CODE
update_tree	proc	near

	push	ax	; save registers
	push	bx
	push	cx
	push	dx
	push	es

	mov	swapping_function, cx	; save function address
	mov	es, tree_seg

UTreeLoop:
	push	ax	; location to update - save for later
	mov	bx, ax	; bl <- location to update
	shl	bx, 1
	shl	bx, 1	; bx points to weight pointer/ bit word

	mov	dx, es:[bx+2]	; save this nodes parent pointer

; increment weight of this node

	mov	ax, es:[bx]	; ax <- weight pointer/ bit
	and	ax, NOT_MSB	; mask out the bit, leaving the weight pointer

	mov	es, weight_list_seg
	mov	bx, ax
	shl	bx, 1
	shl	bx, 1	; es:[bx] is the weight of this node
	inc	word ptr es:[bx]	; increment it
	mov	ax, es:[bx]	; ax <- weight of this node
	cmp	ax, 0ffffh	; nodes too large?
	jb	DontClear

	call	scale_weights

DontClear:

; check if this node has become heavier than nodes above it

	or	bx, bx	; if weight pointer is zero, this is the heaviest already
	jz	DoneWUpdate
	cmp	ax, es:[bx-WLISTLEN]; look at previous weight
	jbe	DoneWUpdate	; if less than or equal to previous weight, no updating

; it has. need to swap with another node
; scan for the next node of equal or greater weight

	mov	cx, bx	; this nodes weight pointer

UTreeScan:
	sub	bx, WLISTLEN	; weight above it
	cmp	bx, 0	; if weight pointer is less than zero, reached heaviest
	jl	DoneScan
	cmp	ax, es:[bx]	; look at new weight
	jbe	DoneScan 	; if less than or equal to new weight reached the end
	jmp	short	UTreeScan	; keep looking

; swap these two nodes. 

DoneScan:
	add	bx, WLISTLEN	; now es:bx points to swapping node weight

; first check if swap is current's parent. if it is, cancel swap

	cmp	dx, es:[bx+2]	; current's parent == swap address?
	je	DoneWUpdate	; cancel swap

	call	swapping_function

DoneWUpdate:

	mov	es, tree_seg	; is this the last node in the tree?
	pop	bx	; recover tree location
	shl	bx, 1
	shl	bx, 1	; es:[bx+2] is parent pointer

	mov	ax, es:[bx+2]	; ax <- parent of this node
	cmp	ax, ROOT	; if it's the root, we're done
	je	DoneUTreeLoop
	jmp	UTreeLoop	; walk up another node

DoneUTreeLoop:

	pop	es	; restore the registers
	pop	dx
	pop	cx
	pop	bx
	pop	ax

	ret

update_tree	endp

;---------------------------------------------------------------------------
; load_buffer
; read filebufsiz bytes from a file into an input buffer.
; on entry:
;	dx:0 - pointer to input buffer
;	bx - file handle
;	si - pointer to current location in buffer
; on exit:
;	si - cleared
;	dx - actual bytes read
;	carry flag set if no more data available
;---------------------------------------------------------------------------
load_buffer	proc	near

	push	ax	; save registers
	push	cx

	push	ds

	mov	ah, 3Fh	; dos function number
	mov	cx, FILEBUFSIZ	; number of bytes to read
	mov	ds, dx	; segment of buffer
	xor	dx, dx	; offset of buffer
	int	21h	; make call

	pop	ds	; restore data segment

	or	ax, ax	; any bytes read?
	jz	NoneLeft
	xor	si, si	; clear si ( and carry flag)
	jmp	short	LBufExit

NoneLeft:
	xor	si, si	; clear current offset
	stc	; set carry

LBufExit:
	mov	dx, ax	; return value (number of bytes)
	pop	cx
	pop	ax

	ret

load_buffer	endp

;---------------------------------------------------------------------------
; dump_buffer
; dump the output buffer to the output file handle. clear the local byte
; counter and update the total byte counter.
; this routine does no error checking.
; on entry:
;	es:0 - pointer to start of buffer
;	di - length of buffer
;	bx - output file handle
; on exit:
; 	di - current offset in output buffer (cleared)
;	global totalbytes updated
;---------------------------------------------------------------------------
dump_buffer	proc	near

	push	ax	; save all the registers
	push	cx
	push	dx

	push	ds

	mov	ah, 40h	; dos function number
	mov	cx, di	; length of buffer
	xor	dx, dx	; offset of buffer
	push	es
	pop	ds	; segment of buffer

	int	21h	; make call

	pop	ds	; data segment
	add	ax, totalbytes	; update total bytes written
	adc	totalbytes+2, 0
	mov	totalbytes, ax

	xor	di, di	; clear di

	pop	dx	; save all the registers
	pop	cx
	pop	ax

	ret
dump_buffer	endp

;---------------------------------------------------------------------------
; add_daughter
; add a daughter node given a location, a parent, and a branch.
; assign parent's location to this location's parent pointer
; assign a weight and position in the weight list to this location
; initialize this node's daughter pointer's to -1
; assign this location to the parents daughter pointer
; (this is just the compressors add-node routine with an additional step to
; account for daughter pointers.)
; NOTE: this routine calls add_node.
;
; on entry:
;	lsb of al - branch on which to attach this node
;	bx - location of this node
;	dx - parent pointer
;	requires access to globals child_ptr_seg, tree_seg, weight_list_seg
; 	and next_weight
; on exit:
;	dx, bx preserved
;---------------------------------------------------------------------------
add_daughter	proc	near

	push	es	; save registers
	push	dx
	push	bx

	push	ax	; save the branch
	push	dx	; save the parent pointer

; call the compressor's add-node routine to handle weight and parent pointer updates

	call	add_node

; now update parent's daughter pointer in tree list

	mov	es, child_ptr_seg
	pop	bx	; parent location
	shl	bx, 1
	shl	bx, 1	; es:[bx] is parent's zero-daughter pointer

	pop	dx	; branch to update
	and	dl, 1	; if zero, do this one
	jz	ZeroChild
	inc	bx	; if one, do the other one
	inc	bx

ZeroChild:
	mov	es:[bx], ax	; store the proper daughter pointer

; now update this node's daughter pointers

	mov	bx, ax
	shl	bx, 1
	shl	bx, 1	; es:[bx] is this node's zero-daughter pointer
	mov	word ptr es:[bx], -1	; update it
	mov	word ptr es:[bx+2], -1	; and the one-daughter pointer

; done adding

	pop	bx	; restore  the registers
	pop	dx
	pop	es
	ret

add_daughter	endp

;---------------------------------------------------------------------------
; incode
; read in a variable length huffman code from a buffer
; walk down the huffman tree, starting at the root.
; start at the root node, and read in a bit. Then jump to location at
; the daughter pointer specified by the bit. Continue reading and jumping
; until daughter pointers for a given node are -1, which indicates leaf.
; return the location of this node as the code.
; on entry:
;	es:si - pointer to input buffer
;	global in_bitshift stores bit offset
; requires access to global child_ptr_seg
; on exit:
;	es:si - points to new input buffer location
;	ax - code
;	global in_bitshift updated
;---------------------------------------------------------------------------
incode	proc	near

	push	es

; read in a byte from the buffer

	mov	al, es:[si]	; al contains the byte from the buffer
	mov	cl, in_bitshift
	shr	al, cl	; al contains the bits in proper position

;point at the root node in the daughter pointer segment

	mov	es, child_ptr_seg
	mov	bx, ROOT
	shl	bx, 1
	shl	bx, 1	; es: [bx] has zero-daughter pointer

; use ds as pointer to input

	push	ds	; save current data seg
	mov	ds, input_seg

ICLoop:
	test	al, 1	; if lsb is set, jump to one-daughter
	jz	JumpToChild
	inc	bx
	inc	bx	; now es:[bx] has one-daughter pointer

JumpToChild:
	mov	bx, es:[bx]	; bx is child
	shl	bx,1
	shl	bx,1	; es:[bx] points to next zero-daughter

; determine if this is a leaf (both daughter pointers are -1)
	cmp	word ptr es:[bx], -1	; check zero child
	jne	NotLeaf
	cmp	word ptr es:[bx+2], -1	; and one child
	jne	NotLeaf

; this is a leaf. save its location 
	shr	bx, 1
	shr	bx, 1
	mov	dx, bx	; return value is in dx
	jmp	short	IC_DoneLoop	; done searching

NotLeaf:
; look at the next bit and jump again

	inc	cl	; update the bit counter
	and	cl, THREE_LSBS	; if cl = 8, cl <- 0
	jz	IC_RELOAD	; need to reload the buffer byte

	shr	al,1	; now lsb of al has the next bit
	jmp	short	ICLoop

IC_Reload:
	inc	si	; point at the next byte
	mov	al, [si]	; get another byte
	jmp	short	ICLoop

IC_DoneLoop:
; done with the loop. account for the last bit read

	pop	ds	; restore current data segment

	inc	cl
	and	cl, THREE_LSBS	; if cl = 8, cl <- 0
	jnz	IC_NoInc
	inc	si	; point si at the next byte

IC_NoInc:
	mov	in_bitshift, cl	; update in_bitshift
	mov	ax, bx	; return value

	pop	es
	ret

incode	endp

;---------------------------------------------------------------------------
; inbyte
; read in one byte from a bit buffer
; on entry:
;	es:si - pointer to byte in buffer
;	global in_bitshift stores bit offset
; on exit:
;	es:si - pointer to new input buffer location
;	al - code
;	ah - cleared
;	global in_bitshift updated
;---------------------------------------------------------------------------
inbyte	proc	near

	push	cx

	mov	ax, es:[si]	; read in a word
	mov	cl, in_bitshift	; prepare to shift
	shr	ax, cl	; shift 
	xor	ah, ah	; clear the high byte
	inc	si	; point to the next byte

	pop	cx
	ret

inbyte	endp
;---------------------------------------------------------------------------
; bit_load_buffer
; load a buffer being used for variable length input
; The last byte in the buffer may or may not be partially
; read when this routine is called.
; RESERVE bytes are reserved at the end of the buffer. When the
; caller overruns the buffer, it calls this routine again. This code
; responds, again reserving RESERVE bytes. When the file is empty, it makes the
; RESERVE bytes available; subsequent calls will fail.
; on entry:
;	dx:si points to last addressed byte -- may be partially read
;	bx is the input file handle
; on exit:
;	dx = size of the input buffer
;	si = 0
;---------------------------------------------------------------------------
.DATA

	blb_empty	db	0	; is the file empty?
	actual_len	dw	?	; how long is it, really?

.CODE
bit_load_buffer	proc	near

	cmp	blb_empty,0	; emptied already?
	je	DoBLB
	stc	; set the carry flag and return
	ret

DoBLB:
	push	ax	; save the registers
	push	cx
	push	es

	mov	es, dx	; input buffer

	push	bx	; save the filehandle

; move actual_len-si bytes to the front of the buffer
; (if si is zero or equals actual_len, skip this step)

	xor	dx, dx	; offset to read in to 

	mov	ax, si	; see if si is zero
	or	ax, ax
	jz	SkipMove

	sub	ax, actual_len	; subtract si from actual len
	neg	ax	; leave difference in cx
	mov	cx, ax
	xor	bx, bx	; clear bx, the loop counter

	jcxz	SkipMove

BLB0:
	mov	al, es:[si]
	mov	es:[bx], al
	inc	bx
	inc	si
	loop	BLB0

	mov	dx, bx	; offset to read in to (spare the bytes we moved)

SkipMove:

	pop	bx	; filehandle
	push	ds	; save the data segment

	push	es
	pop	ds	; ds <- input buffer ptr

	mov	ah, 3Fh	; dos function number
	mov	cx, FILEBUFSIZ	; number of bytes to read
	sub	cx, dx	; fill in only what we don't have
	int	21h	; make call

	pop	ds	; restore data segment

	xor	si, si	; clear si
	add	ax, dx	; add in the reserved bytes
	cmp	ax, RESERVE	; if ax less than RESERVE, file is empty
	ja	NotEmpty
	inc	blb_empty	; set empty flag and don't reserve bytes
	jmp	BLB_IsEmpty

NotEmpty:
	mov	actual_len, ax
	sub	ax, RESERVE	; cheat bytes

BLB_IsEmpty:
	clc
	mov	dx, ax
	pop	es
	pop	cx
	pop	ax
	ret

bit_load_buffer	endp

;---------------------------------------------------------------------------
; dump_bit_buffer
; dump the output buffer to the output file handle. clear the local byte
; counter and update the total byte counter.
;
; check to see if the buffer has data. if no data, return.
; if data, dump up to logical end. save the remaining bits.
; this routine calls dump_buffer.
; on entry:
;	es:0 - pointer to start of buffer
;	ax - logical end of buffer
;	di - filled length
;	bx - output file handle
;	requires access to global out_bitshift
; on exit:
; 	di - current offset in output buffer
;	global totalbytes updated
;---------------------------------------------------------------------------
dump_bit_buffer	proc	near

	cmp	out_bitshift, 0
	jnz	SomeData
	or	di, di
	jnz	SomeData
	ret	; no data, so return

SomeData:
	push	cx
	push	si

	push	ds

	push	di	; save the filled length
	mov	di, ax	; number of bytes to dump
	call	dump_buffer
	
	pop	cx	; the filled length
	sub	cx, ax	; number of bytes left
	inc	cx	; count last byte
	mov	si, ax	; start location

	push	es
	pop	ds	; ds:si is start of tail to move

	rep	movsb	; move the tail
	dec	di	; point back at last byte moved

	pop	ds	; restore the data segment

	mov	cl, out_bitshift
	mov	al, 0ffh	; set all bits
	shl	al, cl
	not	al	; prepare mask for last byte

	and	es:[di], al	; mask it

	pop	si	; restore registers
	pop	cx
	ret
dump_bit_buffer	endp

;---------------------------------------------------------------------------
; c_swap
; compression tree node swapper. swaps:
; in the weight list segment -- weight, and tree ptr
; in the tree segment -- parent pointer, and weight ptr bit
; on entry:
;	ax - current weight
;	bx - pointer to swap node in the weight list
;	cx - current weight pointer
;	es - weight list
; on exit
;	nodes swapped, registers not preserved
;---------------------------------------------------------------------------
c_swap	proc	near
; swap the two words in the weight list between swap and current

	xchg	ax, es:[bx]	; swap weights
	xchg	bx, cx	; bx <- current weight pointer
	mov	es:[bx], ax	; weights swapped
	inc	bx
	inc	bx	; bx points to tree pointer
	mov	ax, es:[bx]	; ax <- current tree pointer
	push	ax	; save it
	xchg	bx, cx	; bx <- swap weight pointer, cx current tree
	inc	bx
	inc	bx	; bx point to swap tree ptr
	xchg	ax, es:[bx]	; swap tree ptrs
	xchg	bx, cx	; bx current tree ptr, cx swap tree ptr
	mov	es:[bx], ax	; tree ptrs swapped
	pop	bx	; recover current tree ptr

; ax has current tree node ptr. bx has swap node ptr
; swap the two words in the tree between current, and swap

	mov	es, tree_seg	; point at the tree segment

	shl	ax, 1
	shl	ax, 1	; ax <- pointer to weight list ptr/bit (swap)
	shl	bx, 1
	shl	bx, 1	; bx <- pointer to weight list ptr/bit (current)
	mov	cx, ax	; cx pointer to swap stuff
	mov	ax, es:[bx]	; ax <- current weight list ptr/bit
	xchg	bx, cx	; bx points to swap stuff, cx to current
	xchg	ax, es:[bx]	; ax <- swap weight list ptr/bit
	xchg	bx, cx	; bx points to current, cx to swap stuff
	mov	es:[bx], ax	; weight list ptr/bits swapped.

	mov	ax, es:[bx+2]	; ax <- parent ptr (current)
	xchg	bx, cx	; bx <- pointer to parent ptr(swap)
	xchg	ax, es:[bx+2]	; ax <- parent ptr (swap)
	mov	bx, cx	; bx <- pointer to parent ptr(current)
	mov	es:[bx+2], ax	; swapping done

	ret
c_swap	endp

;---------------------------------------------------------------------------
; e_swap
; expansion tree node swapper. swaps:
; in the weight list segment -- weight, and tree ptr
; in the tree segment -- parent pointer, and weight ptr bit
; also updates parent's daughter pointers
; on entry:
;	ax - current weight
;	bx - pointer to swap node in the weight list
;	cx - current weight pointer
;	es - weight list
; on exit
;	nodes swapped, registers not preserved
;---------------------------------------------------------------------------
e_swap	proc	near

; swap the two words in the weight list between swap and current

	xchg	ax, es:[bx]	; swap weights
	xchg	bx, cx	; bx <- current weight pointer
	mov	es:[bx], ax	; weights swapped
	inc	bx
	inc	bx	; bx points to tree pointer
	mov	ax, es:[bx]	; ax <- current tree pointer
	push	ax	; save it
	xchg	bx, cx	; bx <- swap weight pointer, cx current tree
	inc	bx
	inc	bx	; bx point to swap tree ptr
	xchg	ax, es:[bx]	; swap tree ptrs
	xchg	bx, cx	; bx current tree ptr, cx swap tree ptr
	mov	es:[bx], ax	; tree ptrs swapped
	pop	bx	; recover current tree ptr

; ax has current tree node ptr. bx has swap node ptr
; swap the two words in the tree between current, and swap

	mov	es, tree_seg	; point at the tree segment

	shl	ax, 1
	shl	ax, 1	; ax <- pointer to weight list ptr/bit (swap)
	shl	bx, 1
	shl	bx, 1	; bx <- pointer to weight list ptr/bit (current)

	push	ax	; save these for updating parents
	push	bx

	mov	cx, ax	; cx pointer to swap stuff
	mov	ax, es:[bx]	; ax <- current weight list ptr/bit
	xchg	bx, cx	; bx points to swap stuff, cx to current
	xchg	ax, es:[bx]	; ax <- swap weight list ptr/bit
	xchg	bx, cx	; bx points to current, cx to swap stuff
	mov	es:[bx], ax	; weight list ptr/bits swapped.

	mov	ax, es:[bx+2]	; ax <- parent ptr (current)
	xchg	bx, cx	; bx <- pointer to parent ptr(swap)
	xchg	ax, es:[bx+2]	; ax <- parent ptr (swap)
	mov	bx, cx	; bx <- pointer to parent ptr(current)
	mov	es:[bx+2], ax	; swapping done

; update the parent's daughter pointers

	pop	bx	; restore the node pointers
	pop	ax

; ax has current tree node ptr. bx has swap node ptr

	mov	cx, es:[bx]	; msb of cx is swap branch
	xchg	bx, ax
	mov	dx, es:[bx]	; msb of dx is current branch

	mov	bx, es:[bx+2]	; bx points to current parent
	xchg	ax, bx	; ax points to current parent
	mov	bx, es:[bx+2]	; ax points to current parent, bx to swap parent

	mov	es, child_ptr_seg
	shl	bx, 1
	shl	bx, 1	; es:[bx] is swap parents zero-daughter pointer
	shl	ax, 1
	shl	ax, 1	; es:[ax] is current parents zero-daughter pointer

	test	dx, MSB_ONLY ; which branch for current?
	jz	CurZChild
	inc	ax
	inc	ax	; now es:[ax] is appropriate daughter pointer

CurZChild:
	test	cx, MSB_ONLY ; which branch for swap?
	jz	SwapZChild
	inc	bx
	inc	bx	; now es:[bx] is appropriate daughter pointer

SwapZChild:
	mov	dx, es:[bx]	; dx is swap daughter pointer
	xchg	ax, bx	; es:[bx] is current daughter pointer
	xchg	dx, es:[bx]	; current daughter pointer replaced
	mov	bx, ax	; es:[bx] is swap daughter pointer
	mov	es:[bx], dx	; swap daughter pointer replaced

	ret
e_swap	endp

;---------------------------------------------------------------------------
; pop_bits
; pop a bitstack to the output buffer
; on entry:
; 	ds:si - point to top of bitstack
; 	cx - number of bits to pop
; 	bx - file handle
; 	dx - length of output buffer
; 	es:di - pointer to output buffer location
; 	bitshift is in out_bitshift global
; on exit:
;	es:di - points to current output buffer location
;	out_bitshift updated
;---------------------------------------------------------------------------
pop_bits	proc	near

	push	bx	; save some registers
	push	dx

; first, if we're getting near the end, dump
; the buffer.

	cmp	di, dx	; see if this is the end of the buffer
	jbe	PS_DontDump

	mov	ax, dx
	call	dump_bit_buffer

PS_DontDump:

; read in the first word off the stack

	mov	ax, [si]

; fixup first byte of stack -- shift low bits to high bits in last byte
; to make stack contiguous

	mov	bx, cx	; bx <- number of bits to pop
	and	cx, THREE_LSBS	; cx <- leftover bits
	jz	DoneFixup	; no fixing required
	sub	cx, 8
	neg	cx	; cx is number of bits to shift left
	shl	al, cl	; shift them
	mov	[si], al	; store them

DoneFixup:
	mov	dh, out_bitshift
	mov	dl, es:[di]	; dl is last output byte
	or	dh, dh	; does this bit contain valid data?
	jne	ByteOK	; yup
	xor	dl, dl	; no, clear it

ByteOK:
	add	bl, dh	; bx <- total number of bits to write
	adc	bh, 0
	sub	cl, dh	; which way to shift stack to meet buffer?
	ja	ShiftRight
	jb	ShiftLeft

; no shifting needed
	or	dl, al	; update the output byte

NSCont:
	mov	es:[di], dl	; send it out to the buffer
	inc	di

	cmp	bx, 8	; are we through yet?
	jz	DonePop	; yes
	jb	DecOne	; need to point back to current byte, then done

; more bytes to write
	sub	bx, 8	; update the bit counter
	inc	si	; point to next byte
	mov	dl, [si]	; load the next byte
	jmp	NSCont

ShiftRight:
	shr	ax, cl	; shift byte by (stack offset-buffer offset)
	or	dl, al	; update the output byte

SRCont:
	mov	es:[di], dl	; send it out to the buffer
	inc	di

	cmp	bx, 8	; are we through yet?
	jz	DonePop	; yes
	jb	DecOne	; need to point back to current byte, then done

; more bytes to write
	sub	bx, 8	; update the bit counter
	inc	si	; point to next byte
	mov	dx, [si]	; load the next word
	shr	dx, cl	; shift it
	jmp	SRCont

ShiftLeft:
	neg	cx	; shift bytes by (output offset - stack offset)
	shl	ax, cl	; shift em
	or	dl, al	; update the output byte

SLCont:
	mov	es:[di], dl	; send it out to the buffer
	inc	di

	cmp	bx, 8	; are we through yet?
	jz	DonePop	; yes
	jb	DecOne	; need to point back to current byte, then done

; more bytes to write
	sub	bx, 8	; update the bit counter
	mov	dx, [si]	; load the next word
	shl	dx, cl	; shift it
	mov	dl, dh	; need to send out high byte
	inc	si	; point to next byte
	jmp	SLCont

DecOne:
	dec	di	; point back at current byte

DonePop:
	and	bl, THREE_LSBS	; if bitshift = 8, bitshift should be zero
	mov	out_bitshift, bl	; update it

	pop	dx	; restore the registers
	pop	bx

	ret

pop_bits	endp

;---------------------------------------------------------------------------
; scale weights
; scale the weights in the weight list segment. this routine is
; called if a weight approaches the capacity of an int
; on entry:
;	es - points to weight list segment
; on exit:
;	all registers preserved.
;---------------------------------------------------------------------------
scale_weights	proc	near
	push	ax	; save registers
	push	bx
	push	cx
	push	dx
	push	di
	push	si
	push	ds

; find the bottom of the list

	mov	bx, next_weight
	dec	bx
	mov	cx, bx	; cx is the number of nodes to scale
	dec	bx	; bx is the offset of the next-to-last entry
	shl	bx,1
	shl	bx,1	; es:bx points to weight of next-to-last entry
	mov	ds, tree_seg

Skip1s:
	cmp	word ptr es:[bx], 1	; if these are one weight leaves, skip them
	ja	ClearLoop
	sub	bx, WLISTLEN	; move to the next leaf
	loop	Skip1s
	jmp	EndLoop

ClearLoop:
	cmp	byte ptr es:[bx+(WLISTLEN-1)], 0 ; if these are non-leaf nodes, skip them.
	jz	ClearLeaf
	sub	bx, WLISTLEN
	loop	ClearLoop
	jmp	EndLoop

ClearLeaf:
; assign this node the weight of the node below it. Then
; walk up the tree, subtracting the difference from all the parents

	mov	ax, es:[bx]	; ax <- weight of this node
	mov	dx, es:[bx+WLISTLEN]; weight of the node below this
	mov	es:[bx], dx	; assign the low weight to this weight 
	sub	ax, dx	; ax <- difference from old weight to new
	mov	di, es:[bx+2]	; di <- tree ptr
	push	cx
	mov	cl, 2	; use for shifting

ClimbLoop:
	shl	di, cl	; multiply for word access
	mov	di, [di+2]	; di <- parent location
	mov	si, di	; save parent location
	cmp	di, ROOT	; have we reached the top of the tree
	je	EndClimb	; yes
	shl	di, cl	; no -- move to parent
	mov	di, [di]	; di <- weight offset
	shl	di, cl	; es:[di] is weight
	mov	dx, es:[di]	; dx <- parent's weight
	sub	dx, ax	; update weight
	mov	es:[di], dx	; store it
	mov	di, si	; recover this node's location
	jmp	short	ClimbLoop

EndClimb:
	pop	cx
	sub	bx, WLISTLEN	; point bx at the next weight
	loop	ClearLoop	; continue clearing

EndLoop:
	pop	ds	; restore registers
	pop	si
	pop	di
	pop	dx
	pop	cx
	pop	bx
	pop	ax

	ret

scale_weights	endp
	END
