; lzw.asm
;
; LZW compression routines - version 1.0
; 808x assembly language, TASM compatible
; Steve Apiki, 10/90
;
; This code demonstrates algorithms discussed in the March '91 issue of BYTE.
;
;     *********************************************************
;     NOTE: This source code is based, in part, on an algorithm
;     described in UNYSIS Corporation's patent No. 4,558,302.
;     McGraw-Hill cautions users that the implementation of this
;     source code may infringe upon the patent rights of UNYSIS.
;     McGraw-Hill disclaims any liability for use of the source
;     code.
;     *********************************************************
;
;
; Code consists of the following blocks:
;
;	_lzw_compress	- c-language callable LZW compression routine
;	_lzw_expand	- c-language callable LZW decompression routine
;
; compression routines:
;	add_code	- adds a code to the string table
;	lookup	- lookup a code in the string table
;	compress_clear	- clear old leaves from the compression table
;	outbits	- write a variable length code to a buffer
;	load_buffer	- load a buffer from a file
;
; expansion routines:
;	add_entry	- add an entry to the expansion string table
;	translate	- translate a code to a string
;	popstring	- pop the output string to a buffer
;	expand_clear	- clear old leaves from the expansion table
;	inbits	- read a variable length code from a buffer
;	bit_load_buffer	- load a buffer of variable length codes from disk
;
; support routines:
;	minbits	- return the minimum bits required to represent a code
;	dump_buffer	- dump a buffer to a file
;
DOSSEG
.MODEL COMPACT

public  _lzw_compress, _lzw_expand

; equates

FILEBUFSIZ	equ	32767	; size of in and out file buffers
HASHSIZE	equ	12253	; the size of the tables which must be hashed
INDXSIZE	equ	7936	; the size of the indexed areas of the table
NUMCHARS	equ	256	; the number of literal chars
MINBITLEN	equ	9	; minimum bits
MAXBITLEN	equ	13	; maximum number of bits
LASTCODE	equ	8191	; the longest allowable code in a 13-bit table
CODELISTSIZE	equ	4096	; max 13-bit codes
CODEBYTES	equ	2	; number of bytes per entry
STRINGLEN	equ	1024	; for the expansion string
RESERVE	equ	4	; number of bytes to reserve for variable length input
COBUFLEN	equ	FILEBUFSIZ-(MAXBITLEN+1) ; the logical output buffer length - compressor
EOBUFLEN	equ	FILEBUFSIZ-STRINGLEN ; the logical output buffer length - expander
EMPTY	equ	7fffh	; the value of an unused basecode (top bit is reserved, hence the odd value)
MSB_ONLY	equ	1000000000000000b	; mask
NOT_MSB	equ	0111111111111111b	; mask
FIVE_LSBS	equ	0000000000011111b	; mask

.DATA

; variables in common to these routines

	output_seg	dw	? 	; segment of output buffer
	input_seg	dw	?	; segment of input buffer
	basecode_seg	dw	?	; segment of the basecode table
	char_seg	dw	?	; segment of the last char table
	av_codelist_seg	dw	?	; segment of available code list
	newcode_seg	dw	?	; segment of the code array
	cndx_seg	dw	?	; segment of the cross-index list

; this is global to avoid a lot of overhead in outbits

	totalbytes	dw	0	; bytes we've written, total
	dw	0	; it's a double word

;-------------------------------------------------------
; lzw_compress
; long lzw_compress( int infh, int outfh);
; c-callable lzw compressor
; compresses using lzw algorithm, 9-13 bit variable length codes, and
; oldest leaf table clearing.
; returns size of output file (bytes)
; returns zero on error
; 
; infh -- file handle of input file (must be opened for binary reads)
; outfh -- file handle of output file (must be opened for binary writes)
;-------------------------------------------------------
; lzw_compress data

	c_ibuf_end	dw	?	; the end of the input buffer

	c_avc_off	dw	0	; offset of next available code
	c_num_avc	dw	0	; the number of codes in this list

	c_next_c	dw	NUMCHARS	; the next available code
	c_cur_bits	db	MINBITLEN	; the current code size
	c_bshift	db	0	; the current bit shift in the output buffer

	c_ifh	dw	?	; in file handle
	c_ofh	dw	?	; out file handle

	c_tab_full	db	0	; set if table has ever been filled

.CODE
_lzw_compress	proc	near

;
	dummy0	struc	; for accessing variables on stack
	bptr0	dw	?
	raddr0	dw	?
	infh	dw	?
	outfh	dw	?
	dummy0	ends

lzc	equ	[bp-bptr0]

	push 	bp	; stack frame manipulations
	mov 	bp,sp
	sub	sp,bptr0

	push	si
	push	di	; preserve index registers for turbo c.

; initialize local variables

	mov	c_next_c, NUMCHARS
	mov	c_cur_bits, MINBITLEN
	mov	c_avc_off, 0
	mov	c_num_avc, 0
	mov	c_bshift, 0
	mov	c_tab_full, 0

; clear return value

	xor	ax, ax
	mov	totalbytes, ax
	mov	totalbytes+2, ax

; allocate memory for file buffers and tables 

	xor	cx, cx	; keep track of what we allocate
	mov	bx, ((HASHSIZE*CODEBYTES)/16)+1

	mov	ah, 48h
	int	21h
	jc	C_MemError
	inc	cl
	mov	basecode_seg, ax	; basecode table

	mov	ah, 48h
	int	21h
	jc	C_MemError
	inc	cl
	mov	newcode_seg, ax	; code table

	shr	bx, 1	; char_seg needs only bytes, not words
	inc	bx

	mov	ah, 48h
	int	21h
	jc	C_MemError
	inc	cl
	mov	char_seg, ax	; character code table

	mov	bx, (FILEBUFSIZ/16)+1
	mov	ah, 48h
	int	21h
	jc	C_MemError
	inc	cl
	mov	input_seg, ax	; input buffer

	mov	ah, 48h
	int	21h
	jc	C_MemError
	inc	cl
	mov	output_seg, ax	; output buffer

	mov	bx, ((CODELISTSIZE*CODEBYTES)/16)+1	; for 13-bit codes
	mov	ah, 48h
	int	21h
	jc	C_MemError
	inc	cl
	mov	av_codelist_seg, ax	; list of available codes

	mov	ah, 48h
	mov	bx, ((INDXSIZE*CODEBYTES)/16)+1
	int	21h
	jc	C_MemError
	mov	cndx_seg, ax	; cross-index

	jmp	short	ClearTables

C_MemError:
	cmp	cx, 0	; how many blocks to return?
	jne	CMEr0
	jmp	LZC_ErrorExit0

CMEr0:
	cmp	cx, 1
	jne	CMEr1
	jmp	LZC_ErrorExit1

CMEr1:
	cmp	cx, 2
	jne	CMEr2
	jmp	LZC_ErrorExit2

CMEr2:
	cmp	cx, 3
	jne	CMEr3
	jmp	LZC_ErrorExit3

CMEr3:
	cmp	cx, 4
	jne	CMEr4
	jmp	LZC_ErrorExit4

CMEr4:
	cmp	cx, 5
	jne	CMEr5
	jmp	LZC_ErrorExit5

CMer5:
	jmp	LZC_ErrorExit6


ClearTables:
; set all codes to -1 in order to clear
	cld
	mov	es, newcode_seg	; segment for the code tables
	xor	di, di	; offset in the segment is zero
	mov	ax, -1
	mov	cx, HASHSIZE
rep	stosw	; clear HASHSIZE words

; and set all available codes to -1 to clear

	mov	es, av_codelist_seg
	xor	di, di
	mov	ax, -1
	mov	cx, CODELISTSIZE
rep	stosw	; clear CODELISTSIZE words

; set up buffers and file handles

	mov	ax, lzc.infh	; save local copies of file handles
	mov	c_ifh, ax	; save input file handle
	mov	ax, lzc.outfh
	mov	c_ofh, ax	; and output file handle

	xor	si, si	; input buffer offset
	xor	di, di	; output buffer offset

	mov	dx, input_seg	; load the input buffer
	mov	bx, c_ifh	; input handle
	call	load_buffer
	mov	c_ibuf_end, dx	; the size of the input buffer

	mov	dx, ds	; dx contains current data segment
	mov	ds, input_seg	; ds:si points to inbuffer

; read in a character

	lodsb
	xor	ah, ah
	mov	bx, ax	; bx <- first character, for initial base code

; this is the  main compression loop

CompLoop:
	lodsb	; now al contains the new character
	push	ds
	mov	ds, dx	; ds is back to current data segment
	pop	dx	; and dx has the input buffer

	cmp	si, c_ibuf_end	; check for the end of the buffer
	jbe	Cont0	; if not, continue reading from inbuffer

; if it is, reload the buffer and try again

	push	bx	; save parameters
	push	dx
	mov	bx, c_ifh
	call	load_buffer	; try to reload
	mov	c_ibuf_end, dx	; load_buffer returns the end of the new buffer. save it.
	pop	dx
	pop	bx
	jc	JumpOff	; done with main loop

	push	ds	; save current data
	mov	ds, dx	; point ds back at input seg
	pop	dx	; current data back in dx
	jmp	short	CompLoop	; ok, we're reloaded. read in a real character

JumpOff:
	jmp	EOLoop	; go to end of routine

Cont0:
; look up the code (in al), and the basecode (in bx) in the string table
;
	mov	cx, newcode_seg	; for lookup function
	push	dx	; preserve input buffer segment
	call	lookup	; find code in table. return location in dx
	; if ah=1, dx=found value. if ah=0 dx=available space
	or	ah, ah
	jz	NotFound	; if not in table, go to add routine.

; found string in table. just keep gathering codes and looking until we
; find a string that's _not_ in the table

	mov	bx, dx	; bx is offset.
	mov	es, cx	; load a pointer to the code table
	shl	bx, 1	; double for word access
	mov	bx, es:[bx]	; get the new base code from the table

	mov	dx, ds	; dx contains current data segment
	pop	ds	; ds points to the input buffer

	jmp	short	CompLoop

NotFound:
; found a code that's not in the table. write the last complete code to
; the output buffer

	mov	cl, c_cur_bits	; number of bits to use
	mov	ch, c_bshift	; bit offset in output buffer
	mov	es, output_seg	; es:di -- output buffer

	push	dx
	mov	dx, COBUFLEN	; end of output buffer
	push	ax
	mov	ax, c_ofh

	call	outbits	; output the last code

	pop	ax
	pop	dx
	mov	c_bshift, ch	; remember the last c_bshift

; add basecode + current character to tables

	mov	cx, c_next_c	; open code to add to table
	mov	es, basecode_seg	; where to find the basecode table
	call	add_code	; add a new code

; update next code. this is trivial if table is not full, but not if it is full
; with small files the table will never fill and we can skip the code list part

	or	c_tab_full, 0	; has the table ever been filled?
	jnz	BeenFilled	; yes
	mov	bx, c_next_c	; bx has the current code -- the next one to use on output
	call	minbits	; how many bits in this code?
	mov	c_cur_bits, bl	; these are the current bits
	inc	c_next_c	; increment to next code

	cmp	c_next_c, LASTCODE	; will it be filled next time?
	jb	NxCodeDone
	inc	c_tab_full	; if so, set the flag
	jmp	short NxCodeDone

BeenFilled:
; the table may not be full currently, but we must get our codes from the list
; (that is, we can't just inc it)

	mov	bx, c_avc_off	; how many codes have we used
	cmp	bx, c_num_avc	; how many codes in the table
	jb	NoClear	; if it's not full currently, don't clear

	call	compress_clear	; clear the old leaves
	mov	c_num_avc, bx	; number of new available codes
	xor	bx, bx
	mov	c_avc_off, bx	; new number of used codes is zero

NoClear:
	mov	es, av_codelist_seg
	shl	bx, 1	; double for word access
	mov	dx, es:[bx]	; dx is the next avail code
	mov	c_next_c, dx	; next code counter updated
	inc	c_avc_off	; keep count of our position in the available code list

NxCodeDone:
	xor	ah, ah
	mov	bx, ax	; current character becomes new base code

	mov	dx, ds	; dx contains current data segment
	pop	ds	; ds points to the input buffer
	jmp	CompLoop	; continue compressing

EOLoop:
; main compression loop is done
; need to send out one more code before exiting, however

	mov	cl, c_cur_bits	; number of bits to use
	mov	ch, c_bshift	; bit offset in output buffer
	mov	es, output_seg	; es:di -- output buffer
	mov	dx, COBUFLEN	; end of output buffer

	mov	ax, c_ofh
	call	outbits	; output the last code

; round off for final code, unless bshift is zero

	or	ch, ch	; no partial byte left in the buffer
	jz	NoRounding
	inc	di	; partial byte left, increment buffer length

NoRounding:
; dump the output buffer

	mov	bx, c_ofh
	call	dump_buffer	; send it to the file

; return allocated memory
	mov	ah, 49h
	mov	es, cndx_seg
	int	21h

LZC_ErrorExit6:
	mov	ah, 49h
	mov	es, av_codelist_seg
	int	21h

LZC_ErrorExit5:
	mov	ah, 49h
	mov	es, output_seg
	int	21h

LZC_ErrorExit4:
	mov	ah, 49h
	mov	es, input_seg
	int	21h

LZC_ErrorExit3:
	mov	ah, 49h
	mov	es, char_seg
	int	21h

LZC_ErrorExit2:
	mov	ah, 49h
	mov	es, newcode_seg
	int	21h

LZC_ErrorExit1:
	mov	ah, 49h
	mov	es, basecode_seg
	int	21h

LZC_ErrorExit0:
	mov	ax, totalbytes	; return value
	mov	dx, totalbytes+2	; high order word

	pop	di	; restore the registers.
	pop	si
	pop	bp

    	ret
_lzw_compress	endp

;---------------------------------------------------------------------------
; lzw_expand
; long lzw_expand( int infh, int outfh);
; lzw expander
; returns size of output file (bytes)
; returns zero on error
; 
; infh -- file handle of input file (must be opened for binary reads)
; outfh -- file handle of output file (must be opened for binary writes)
;---------------------------------------------------------------------------
; lzw_expand data

.DATA
	e_ibuf_end	dw	?	; the end of the input buffer

	e_avc_off	dw	0	; offset of next available code
	e_num_avc	dw	0	; the number of codes in this list

	e_next_c	dw	NUMCHARS	; the next available code
	e_curbits	db	MINBITLEN	; the current code size
	e_bshift	db	0	; the current bit shift in the input buffer

	e_ifh	dw	?	; in file handle
	e_ofh	dw	?	; out file handle

	e_string	db	STRINGLEN dup (?)  ; the string for the expansion routine
	e_last	dw	?	; the last code

	e_tab_full	db	0	; table full marker

.CODE
_lzw_expand	proc	near

	dummy1	struc	; for access to variables on stack
	bptr1	dw	?
	raddr1	dw	?
	infh	dw	?
	outfh	dw	?
	dummy1	ends

lze	equ	[bp-bptr1]

	push 	bp	; stack frame manipulations
	mov 	bp,sp
	sub	sp,bptr1

	push	si
	push	di	; preserve index registers for turbo c.

; initialize local variables

	mov	e_next_c, NUMCHARS
	mov	e_curbits, MINBITLEN
	mov	e_avc_off, 0
	mov	e_num_avc, 0
	mov	e_bshift, 0
	mov	e_tab_full, 0

; clear return value

	xor	ax, ax
	mov	totalbytes, ax
	mov	totalbytes+2, ax

; allocate memory for file buffers and tables 

	xor	cx, cx	; keep track
	mov	bx, ((INDXSIZE*CODEBYTES)/16)+1

	mov	ah, 48h
	int	21h
	jc	E_MemError
	inc	cl
	mov	basecode_seg, ax	; basecode table

	shr	bx, 1	; char_seg needs only bytes, not words
	inc	bx

	mov	ah, 48h
	int	21h
	jc	E_MemError
	inc	cl
	mov	char_seg, ax	; character code table

	mov	bx, (FILEBUFSIZ/16)+1
	mov	ah, 48h
	int	21h
	jc	E_MemError
	inc	cl
	mov	input_seg, ax	; input buffer

	mov	ah, 48h
	int	21h
	jc	E_MemError
	inc	cl
	mov	output_seg, ax	; output buffer

	mov	bx, ((CODELISTSIZE*CODEBYTES)/16)+1	; for 13-bit codes
	mov	ah, 48h
	int	21h
	jc	E_MemError
	inc	cl
	mov	av_codelist_seg, ax	; list of available codes

	jmp	short	E_Clear

E_MemError:
	cmp	cx, 0	; how many blocks to return?
	jne	EMEr0
	jmp	LZE_ErrorExit0

EMEr0:
	cmp	cx, 1
	jne	EMEr1
	jmp	LZE_ErrorExit1

EMEr1:
	cmp	cx, 2
	jne	EMEr2
	jmp	LZE_ErrorExit2

EMEr2:
	cmp	cx, 3
	jne	EMEr3
	jmp	LZE_ErrorExit3

EMEr3:
	jmp	LZE_ErrorExit4

E_Clear:
; set all basecodes to EMPTY in order to clear .
	cld
	mov	es, basecode_seg	; segment for the basecode tables
	xor	di, di
	mov	ax, EMPTY
	mov	cx, INDXSIZE
rep	stosw	; clear INDXSIZE words

; set up buffers and file handles

	mov	ax, lze.infh	; save local copies of these
	mov	e_ifh, ax
	mov	ax, lze.outfh
	mov	e_ofh, ax

	xor	si, si	; input buffer offset
	xor	di, di	; output buffer offset

	mov	dx, input_seg	; load the input buffer
	mov	bx, e_ifh	; input handle
	call	bit_load_buffer
	mov	e_ibuf_end, dx	; the size of the input buffer

	mov	es, output_seg	; aim at the output buffer

; read in a character, using variable length

	mov	dx, input_seg	; load the input buffer
	mov	cl, e_curbits	; current number of bits
	mov	ch, e_bshift	; current bit_shift
	call	inbits	; bx <- input code
	mov	e_bshift, ch

	mov	ax, bx
	stosb	; write out this code
	mov	e_last, ax	; now this is the last code used

ExpandLoop:
; read in a code

	mov	cl, e_curbits	; the current variable size in bits
	mov	ch, e_bshift	; the current bitshift
	call 	inbits	; bx <- input code
	mov	e_bshift, ch	; update bitshift

; check for buffer overflow

	cmp	si, e_ibuf_end	; have we reached the end of the buffer?
	jb	E_Cont	; if not, continue reading from inbuffer
	ja	E_Reload
	or	ch, ch	; if si is one past the end, see if si is pointing to empty byte
	jz	E_Cont	; if it is, then this byte is valid, too

E_Reload:
; if empty, reload the buffer -- this time we have to be sensitive to
; bit offsets, however

	push	dx
	push	bx
	mov	bx, e_ifh	; input handle
	call	bit_load_buffer	; try to reload
	mov	e_ibuf_end, dx	; the size of the input buffer
	pop	bx
	pop	dx

	jnc	E_Cont	; okay, reloaded
	jmp	EEOLoop	; done - out of buffer

E_Cont:
; throughout this loop, al will contain the last character, and
; bx contains the code we just read. dx holds the input buffer

; lookup the code to see if it's in the string table

	or	bh, bh	; if bh is zero, this is a literal
	jz	E_InTable	; hence automatically in the table

	mov	es, basecode_seg	; if not literal, is it in the table?
	push	bx
	shl	bx, 1
	cmp	word ptr es:[bx-(CODEBYTES*NUMCHARS)], EMPTY	; look for an EMPTY basecode
	pop	bx
	jne	E_InTable

; the code is not in the table. this is the special case, and
; this is the special case handler

	push	bx	; save the code we just read in
	push	dx	; save the input segment

	push	ax	; save last character
	mov	bx, e_last	; translate last code
	mov	dx, offset e_string	
	call	translate	; translate it - string is in e_string

	mov	bx, e_ofh	; file handle for string popstring
	mov	es, output_seg
	mov	ax, EOBUFLEN	; end of buffer for popstring
	call	popstring	; write out the string. save first char in cl.
	pop	ax	; restore the last character
	stosb	; write out the last character

	pop	dx
	pop	bx	; code we just read is back in bx.

	jmp	short	ExceptionDone

; this is the usual case, where the code is in the table.
E_InTable:

	push	bx
	push	dx	; save input buffer
	mov	dx, offset e_string	; translate the current code (bx) to string
	call	translate

	mov	bx, e_ofh
	mov	es, output_seg
	mov	ax, EOBUFLEN	; end of buffer for popstring
	call	popstring	; write out the string. save first char in cl.

	pop	dx	; restore what we've saved
	pop	bx

ExceptionDone:
	mov	al, cl	; lastchar <- first char from string

; add the next code to the table

	mov	cx, e_next_c	; code to add
	xchg	e_last, bx	; e_last code <- current code ; bx <- last
	call	add_entry	; add the current entry to the table
	; using last (bx) , char (al), and next code (cx)

; need to come up with a new code.  are there still free codes available?

	or	e_tab_full, 0	; has the table been filled?
	jnz	E_BeenFilled
	inc	e_next_c	; if not, we can just increment to next code
	mov	bx, e_next_c	; bx has the new code
	call	minbits	; how many bits in this code?
	mov	e_curbits, bl	; these are the current bits

	cmp	e_next_c, LASTCODE	; will it be filled next time?
	jb	J0
	inc	e_tab_full	; if so, set the flag
J0:
	jmp	ExpandLoop	; if not, just continue

; the table may not be full currently, but we must get our codes from the list
E_BeenFilled:

	mov	cx, e_avc_off	; how many codes have we used
	cmp	cx, e_num_avc	; how many codes in the list
	jb	E_NoClear
	call	expand_clear	; clear the old leaves
	mov	e_num_avc, cx	; number of available codes as returned by clear routine
	xor	cx, cx	; cx is the number of codes used
	mov	e_avc_off, 0	; reset the codes used

E_NoClear:
	mov	es, av_codelist_seg
	xchg	bx, cx	; bx has the offset into the codelist
	shl	bx, 1	; word access
	mov	bx, es:[bx]	; bx has the next available code
	mov	e_next_c, bx	; now e_next_c does
	inc	e_avc_off	; pointer to the next available code
	mov	bx, cx	; bx is back

	jmp	ExpandLoop	; loop again

EEOLoop:
; just about done expanding

; dump the output buffer

	mov	es, output_seg	; the buffer segment
	mov	bx, e_ofh	; the output file handle
	call	dump_buffer

LZE_ErrorExit:
; return allocated memory

	mov	ah, 49h
	mov	es, av_codelist_seg
	int	21h

LZE_ErrorExit4:
	mov	ah, 49h
	mov	es, output_seg
	int	21h

LZE_ErrorExit3:
	mov	ah, 49h
	mov	es, input_seg
	int	21h

LZE_ErrorExit2:
	mov	ah, 49h
	mov	es, char_seg
	int	21h

LZE_ErrorExit1:
	mov	ah, 49h
	mov	es, basecode_seg
	int	21h

LZE_ErrorExit0:
	mov	ax, totalbytes	; return value
	mov	dx, totalbytes+2	; high order word

	pop	di	; restore the registers.
	pop	si
	pop	bp

    	ret
_lzw_expand	endp

;---------------------------------------------------------------------------
; add_code
; add a code to the table. put the basecode in the basecode table,
; the new char in the char table, and the new code in the code table.
; then put the index of this code in the cross index table.
; use the index provided.
; on entry:
;	 al - new char
;	 bx - base code
;	 dx - index into table
;	 es - basecode table segment
;	 cx - code to add
; on exit:
;	preserve al
;	bx, es  _not_ preserved
;---------------------------------------------------------------------------
add_code	proc	near

	xchg	dx, bx	; put the index into bx
	shl	bx, 1	; double for word access
	mov	es:[bx], dx	; store the new basecode

	mov	es, newcode_seg
	mov	es:[bx], cx	; store the new code

	shr	bx, 1	; back to being an index
	mov	es, char_seg
	mov	es:[bx], al	; store the new character

	mov	dx, bx	; dx is the index value

	mov	es, cndx_seg	; aim at the cross-index table
	mov	bx, cx	; bx <- new code, which is index into cross index table
	shl	bx, 1	; double for word access
	mov	es:[bx-(CODEBYTES*NUMCHARS)], dx	; add index to cross index table

	ret
add_code	endp

;-------------------------------------------------------
; lookup
; lookup a base code, new character combination in the string table
; if we find it, return the offset into the table in dx
; if we don't, find the next empty slot, set a flag in ah and return the
; offset of the empty slot in dx.
; on entry:
; 	bx - base code
;	al - new character
; 	cx - code seg
; on exit:
;	al, bx, cx preserved
;	ah - flag <- 1 if code found, 0 if not and dx is an empty slot
;	dx - index into table
;-------------------------------------------------------
lookup	proc	near

	mov	ah, 0	; this is for the flag later on (assume we won't find it)
	push	bx
	push	cx
	push	ds
	push	es
	push	ax	; save all the registers

	mov	es, cx	; es points to the code table

; hash with al, ah, and bx

	mov	cx, ax	; save ax and bx for later
	mov	dx, bx	; cx is add on char, dx is base code
	mov	ah, al	; duplicate 8-bit val to obtain 16-bit val
	and	ah, FIVE_LSBS	; hack off invalid bits
	xor	bx, ax	; xor hash

CheckAgain:
	cmp	bx, HASHSIZE	; make sure we don't overflow
	jb	DoneHash	; if we don't, we're done
	sub	bx, HASHSIZE	; if we do, keep subtracting chunks
	jmp	short	CheckAgain

DoneHash:
	shl	bx, 1	; double for word access
	mov	ax, es:[bx]	; put code from index in ax
	cmp	ax, -1	; is this code available ?
	je	NotInTable	; yes

; did we find it, or is this another code?

	mov	es, basecode_seg	; es points to basecode segment
	mov	ax, es:[bx]	; put basecode from table in ax
	shr	bx, 1	; put bx back to being an index
	cmp	ax, dx	; are the basecodes the same?
	jne	ReHash	; rehash and try again

	mov	es, char_seg	; es points to appended char segment
	mov	al, es:[bx]	; char from table
	cmp	cl, al	; are the chars the same?
	je	InTable	; yes, so we found the code

ReHash:
; rehash, putting the new hash value in bx

	or	dx, dx
	jz	ZeroBasecode
	add	bx, dx	; use basecode as rehash increment value
	dec	bx

ZeroBasecode:
	inc	bx	; use one if basecode is zero
	mov	es, newcode_seg
	jmp	short	CheckAgain

NotInTable:
	shr	bx, 1	; put bx back to being an index
	mov	dx, bx	; put the offset of the empty slot in dx
	pop	ax	; restore the zero flag, and ah
	jmp	short	LUExit

InTable:
	mov	dx, bx	; put the offset of the found slot in dx
	pop	ax
	inc	ah	; set the found flag

LUExit:
	pop	es	; restore the rest of the registers
	pop	ds
	pop	cx
	pop	bx

	ret
lookup	endp

;---------------------------------------------------------------------------
; compress_clear
; clear the compression table by removing old leaves
; on the first pass, remove leaves with 9-bit codes (the oldest).
; on the second, 10-bit, etc., until we reach 13, whereupon we go
; back to 9.
; on entry:
;	ds - current data segment
; on exit:
;	no return value
;	ax, di, si preserved.
;	bx - number of available codes
;---------------------------------------------------------------------------
.DATA
	c_oldest	db	MINBITLEN	; number of bits for oldest leaves

.CODE
compress_clear	proc	near

	push	ax	; save other registers
	push	cx
	push	dx
	push	di
	push	si

; mark all codes as leaves. for leaf designations, borrow the top bit of the
; cross index list

	mov	es, cndx_seg
	xor	di, di	; es:di points to start of cross index list
	mov	cx, INDXSIZE	; number of codes

C_Loop0:
	or	word ptr es:[di], MSB_ONLY	; set top bit
	inc	di
	inc	di	; point to next code
	loop	C_Loop0


; scan code table

	mov	dx, es	; dx holds the cross index list
	mov	es, basecode_seg	; es points to the basecodes
	mov	ax, es	; ax points to the basecodes, too

	push	ds
	mov	ds, newcode_seg
	xor	si, si	; ds:si points to start of newcode list

	mov	cx, HASHSIZE	; number of new codes

C_Loop1:
	cmp	word ptr [si], -1
	jz	C_EOLoop1	; if not valid code, do nothing

	mov	bx, es:[si]	; bx contains a valid base code
	sub	bx, NUMCHARS	; table offset for cross index
	jc	C_EOLoop1	; do nothing if this code has a literal basecode
	shl	bx, 1	; double for word access
	mov	es, dx	; es <- cross index seg
	and	word ptr es:[bx], NOT_MSB ; clear the leaf designation
	mov	es, ax	; es <- basecode_seg

C_EOLoop1:
	inc	si
	inc	si
	loop	C_Loop1	; continue scanning basecodes

	pop	ds	; ds is back to current data

C_TopOfClear:
; now scan the area of interest in the list of leaf markers
; area of interest contains the oldest leaves.

	mov	ax, 1
	mov	cl, c_oldest	; cl <- number of bits in oldest leaves
	dec	cl
	shl	ax, cl
	mov	cx, ax	; cx <- number of codes to look at
	sub	ax, NUMCHARS	; ax is the first code to look at
	mov	si, ax	; si points to the first code

	mov	es, av_codelist_seg
	xor	di, di	; es:di points to list of available codes

	mov	dx, newcode_seg	; dx points to list of new codes

	push	ds
	mov	ds, cndx_seg	; ds:si points to leaf list

C_Loop2:
	lodsw	; ax <- cindx+leaf mark
	test 	ax, MSB_ONLY ; is this marked as a leaf?
	jz	C_EOLoop2	; if not, skip this one
	shl	ax, 1	; double for word access (scraps the high bit, too)
	mov	bx, ax	; bx <- pointer into code table
	push	es
	mov	es, dx	; es points to newcode_seg
	mov	ax, es:[bx]	; ax <- available code
	mov	word ptr es:[bx], -1	; mark code as unused
	pop	es	; es points to available codes
	stosw	; write out an available code

C_EOLoop2:
	loop	C_Loop2	; continue looping

	pop	ds	; ds is back to current data

; update the oldest leaf variable

	inc	c_oldest
	cmp	c_oldest, MAXBITLEN+1
	jne	C_NoWrap
	mov	c_oldest, MINBITLEN	; wraparound

C_NoWrap:
; did this work? there may not have been any free leaves at the selected
; bitlength

	mov	ax, di
	or	ax, ax	; any new codes?
	jz	C_TopOfClear	; nope

	shr	ax, 1	; number of available codes is di/2
	mov	bx, ax

; restore the cross-index segment

	mov	es, cndx_seg
	xor	di, di
	mov	cx, INDXSIZE

C_Loop3:
	and	word ptr es:[di], NOT_MSB
	inc	di
	inc	di
	loop	C_LOOP3

	pop	si	; restore other registers
	pop	di
	pop	dx
	pop	cx
	pop	ax

	ret
compress_clear	endp

;---------------------------------------------------------------------------
; outbits
; output a code, using a fixed number of bits. the code to output comes to this
; routine in bx. number of bits and offsets are also provided.
;
; output buffer bytes look like this:
;	byte 0                                  byte 1
;  MSB                                 LSB
;  c0b7-c0b6-c0b5-c0b4-c0b3-c0b2-c0b1-c0b0     c1b5-c1b4-c1b3-c1b2-c1b1-c1b0-c0b9-c0b8
;
; where c0b1 is code 0, bit 0,  LSB of code 0.
; on entry:
;	 ax - file handle
;	 bx - base code (code to output)
;	 cl - number of bits
;	 ch - current bit offset
;	 dx - length of output buffer
;	 es:di - points to output location
; on exit:
;	 ch - current bit offset
;	 di - current byte offset
;---------------------------------------------------------------------------
outbits	proc	near

	push	bx
	push	cx

; first: if there is no offset, and we're getting near the buffer limit, dump
; the buffer while we have the chance

	or	ch, ch	; is there a file offset?
	jnz	HasOffset	; yes, so can't dump
	cmp	di, dx	; is this the end of the buffer
	jb	DontDump	; no, don't dump

	push	bx	; dump the buffer
	mov	bx, ax	; filehandle in bx
	call	dump_buffer
	pop	bx	; restore bx

DontDump:
; if current shift is zero, clear current byte rather than fetch from buffer

	mov	ah, ch	; move bit shift
	add	ah, cl	; ah <- total number of bits to write 
	xor	al, al	; al is cleared
	jmp	short NBO0End

HasOffset:
	mov	ah, ch	; move bit shift
	add	ah, cl	; ah <- total number of bits to write
	mov	al, es:[di]	; al has the current byte
	xor	dl, dl	; clear dl, our overflow register
	mov	cl, ch	; put the current bit shift in cx
	xor	ch, ch

NBO0:
	shl	bx, 1	; shift bx up by the current shift count
	rcl	dl, 1	; and put overflow in dl
	loop	NBO0

NBO0End:
; now al is a copy of the buffer byte, ready to be ord into
	or	al, bl	; update the current byte
	stosb	; and write it out

	sub	ah, 8	; update the offset
	mov	al, bh	; write out the next byte
	stosb
	cmp	ah, 8	; check if written all we can
	jb	OBExit

	sub	ah, 8	; update the offset
	jz	OBExit	; if even bits, exit without writing another

	mov	al, dl	; write out the next byte
	stosb

OBExit:
	or	ah, ah	; unless shift is zero, we've done
	jz	NoDec	; one too many incs of di.
	dec	di

NoDec:
	pop	cx
	mov	ch, ah	; save current byte offset

	pop	bx
	ret

outbits	endp

;---------------------------------------------------------------------------
; load_buffer
; try to read filebufsiz bytes
; set carry if there is nothing left to read in the file
; on entry:
;	dx:0 - input buffer
;	ds   - data segment
;	bx - file handle
; on exit:
;	si - cleared
;	dx - actual bytes read
;	carry flag set if file is empty
;---------------------------------------------------------------------------
load_buffer	proc	near

	push	ax	; save all the registers
	push	cx

	push	ds

	mov	ah, 3Fh	; dos function number
	mov	cx, FILEBUFSIZ	; number of bytes to read
	mov	ds, dx	; segment of buffer
	xor	dx, dx	; offset of buffer

	int	21h	; make call

	pop	ds	; restore data segment

	or	ax, ax	; any bytes read?
	jz	NoneLeft
	xor	si, si	; clear si ( and carry flag)
	jmp	short	LBufExit

NoneLeft:
	xor	si, si	; clear si
	stc	; set carry flag, nothing left

LBufExit:
	mov	dx, ax	; return value, number of bytes read
	pop	cx	; restore registers
	pop	ax

	ret
load_buffer	endp

;---------------------------------------------------------------------------
; add_entry
; add the current entry to the expansion string table
; straightforward add to an indexed table
; on entry:
;	bx -- base code
;	al -- append char
;	cx -- code to use
; on exit:
;	table updated
;	registers preserved
;---------------------------------------------------------------------------
add_entry	proc	near

	push	es	; save pointer

	xchg	bx, cx	; bx <- code to use,  cx <- base code
	mov	es, char_seg
	mov	es:[bx-NUMCHARS], al ; update append char
	shl	bx, 1	 ; double for word access
	mov	es, basecode_seg
	mov	es:[bx-(CODEBYTES*NUMCHARS)], cx ; update base code
	shr	bx, 1	; bx is back to being index
	xchg	bx, cx	; put all back

	pop	es	; replace pointer

	ret
add_entry	endp

;---------------------------------------------------------------------------
; translate
; translates an lzw code to a string, using the global table. used
; by expander. Start with the code. Push its associated character on a stack.
; jump to code's basecode, and continue pushing until the basecode is literal.
; push that literal basecode, and exit.
; on entry:
;	bx - code to translate
;	dx - address of string
;	requires access to global basecode seg, code seg.
; on exit:
;	cx - length of string
;	dx preserved
;---------------------------------------------------------------------------
translate	proc	near

	or	bh, bh	; if code is literal, we're done
	jnz	NotLiteral
	xchg	dx, bx	; dx <- code, bx <- string address
	mov	[bx], dl	; output the code
	mov	cx, 1	; string length
	xchg	dx, bx	; put all back
	ret	; and exit

NotLiteral:
	push	bx	; save the inputs
	push	dx
	push	es
	push	ds

	mov	es, basecode_seg	; set up the segment pointers
	mov	cx, ds	; cx holds the data segment
	mov	ds, char_seg

	push	dx	; use initial address in string size calculation

KeepTranslating:
; bx is a code in the table

	mov	al, ds:[bx-NUMCHARS]	; al is the character to put in the string
	shl	bx, 1	; basecodes are words
	mov	bx, es:[bx-(CODEBYTES*NUMCHARS)]	; bx has the next basecode

	xchg	dx, bx	; dx <- next base code, bx <- string address
	push	ds
	mov	ds, cx	; ds <-data seg

	mov	ds:[bx], al	; move the character to the string
	inc	bx	; update the string address
	xchg	dx, bx	; dx <- string address, bx <- basecode
	pop	ds	; ds <- char seg
	or	bh, bh	; is the base code a literal?
	jnz	KeepTranslating	; if not, continue looping

; found a literal basecode

	xchg	dx, bx	; dx <- code, bx <- string address
	mov	ds, cx	; ds is back to data
	mov	ds:[bx], dl	; update the string with the last basecode
	inc	bx	; and update the string address

	pop	ax	; original string address
	sub	bx, ax	; length of string

	mov	cx, bx	; return value

	pop	ds	; restore the inputs
	pop	es
	pop	dx
	pop	bx

	ret
translate	endp

;---------------------------------------------------------------------------
; popstring
; pop a string, stored on a stack, to an output buffer. used by lzw expander.
; on entry:
;	ax - end of output buffer
;	bx - output file handle
;	dx - address of string
;	cx - length of string
;	es:di - output
; on return:
;	cl - first char in string
;---------------------------------------------------------------------------
popstring	proc	near

	push	si
	push	ax

	add	dx, cx	; points to string + length
	dec	dx	; points to end of string
	mov	si, dx
	mov	dl, [si]	; return value -- first character in string

PopS0:	; just loop to pop stack to output
	mov	al, [si]
	stosb
	dec	si
	loop	PopS0

; check to see if popping the string made us exceed the output buffer.
; if it does, dump the buffer to a file.
;
	pop	ax	; end of buffer
	cmp	di, ax
	jb	NoDump	; no dumping required

	call	dump_buffer	; dump if necessary

NoDump:
	pop	si
	mov	cl, dl	; return value
	ret

popstring	endp

;---------------------------------------------------------------------------
; expand_clear
; clears the expansion table. Does so by removing leaves from the 
; table when they are more than a few generations old.
; method:
; mark all codes as leaves
; scan basecode table
; for each used basecode
; 	if not literal
;	mark the corresponding code as not-leaf
; scan leaf marks. on the first pass, scan 9-bit area. second pass, 10-bit, etc.
; for each leaf
;	add offset to available code list
;	mark code as unused
; unmark codes
; on entry:
;	ds - current data segment
; on exit:
;	cx - number of available codes
;---------------------------------------------------------------------------
.DATA
	e_oldest	db	MINBITLEN	; number of bits for oldest leaves

.CODE
expand_clear	proc	near

	push	ax	; save other registers
	push	bx
	push	dx
	push	di
	push	si

; mark all codes as leaves. for leaf designations, borrow the top bit of the
; basecode list

	mov	es, basecode_seg
	xor	di, di	; es:di points to start of basecode list
	mov	cx, INDXSIZE	; number of basecodes

E_Loop0:
	or	word ptr es:[di], MSB_ONLY	; set top bit
	inc	di
	inc	di	; point to next code
	loop	E_Loop0

; all codes marked as leaves. see which ones really aren't: scan basecodes

	push	ds
	mov	ds, basecode_seg
	xor	si, si	; ds:si points to start of basecode list
	mov	cx, INDXSIZE	; number of basecodes

E_Loop1:
	lodsw	  ; ax <- basecode+leaf marker
	and	ax, NOT_MSB ; clear the leaf marker, now ax<-basecode
	sub	ax, NUMCHARS	; table offset. code 256 is stored at location 0
	jc	E_EOLoop1	; do nothing if this code has a literal basecode
	shl	ax, 1	; double for word access
	mov	bx, ax	; bx <- pointer to a non-leaf code
	and	word ptr es:[bx], NOT_MSB	; clear the leaf designation
E_EOLoop1:
	loop	E_Loop1	; continue scanning basecodes

	pop	ds	; ds is back to current data

E_TopOfClear:
; now scan the area of interest in the list of leaf markers
; area of interest contains the oldest leaves.

	mov	ax, 1
	mov	cl, e_oldest	; cl <- number of bits in oldest leaves
	dec	cl
	shl	ax, cl
	mov	cx, ax	; cx <- number of codes to look at
	sub	ax, NUMCHARS	; ax is the first code to look at
	mov	si, ax	; si points to the first code

	mov	es, av_codelist_seg
	xor	di, di	; es:di points to list of available codes

	push	ds
	mov	ds, basecode_seg	; ds:si points to leaf list

E_Loop2:
	test 	word ptr [si], MSB_ONLY	; is this marked as a leaf?
	jz	E_EOLoop2	; if not, skip this one
	mov	[si], EMPTY	; if it is a leaf, clear the basecode
	mov	ax, si	; ax <- offset in table of available code
	shr	ax, 1
	add	ax, NUMCHARS	; ax <- available code
	stosw	; store the code in the code list
E_EOLoop2:
	inc	si
	inc	si
	loop	E_Loop2	; continue looping

	pop	ds	; ds is back to current data

; update the oldest leaf variable

	inc	e_oldest
	cmp	e_oldest, MAXBITLEN+1
	jne	E_NoWrap
	mov	e_oldest, MINBITLEN	; wraparound

E_NoWrap:
; did this work? there may not have been any free leaves at the selected
; bitlength

	mov	ax, di
	or	ax, ax	; any new codes?
	jz	E_TopOfClear	; nope
	shr	ax, 1	; number of available codes is di/2

; restore the basecode list to its original form

	mov	es, basecode_seg
	xor	di, di
	mov	cx, INDXSIZE

E_Loop3:
	and	word ptr es:[di], NOT_MSB	; clear top bit
	inc	di
	inc	di	; point to next code
	loop	E_Loop3

	mov	cx, ax	; return value - number of codes

; done - restore

	pop	si	; restore other registers
	pop	di
	pop	dx
	pop	bx
	pop	ax

	ret
expand_clear	endp

;---------------------------------------------------------------------------
; inbits
; input a word, using a variable number of bits
; on entry:
;	dx:si points to input word
;	ch is the current bitshift
;	cl is the number of bits to input
; on exit:
;	si and ch are updated
;	word is in bx
;---------------------------------------------------------------------------
inbits	proc	near

	push	ax	; save contents of ax
	push	dx
	push	ds	; save current data seg

	xor	bh, bh
	mov	bl, ch	; bx contains bit shift
	add	ch, cl	; ch <- total read value

	mov	ds, dx	; ds points to input buffer

	lodsw	; load 16 bits into ax
	sub	ch, 8	; update bits to read counter
	cmp	ch, 8	; have we read enough?
	jb	InRead	; yes, don't read more

	sub	ch, 8	; if exactly one left
	jz	InRead	; don't read more either

	mov	dl,[si]	; more than one left, read another
	inc	si	; point to next bit

InRead:
	xchg	bx, cx	; cx <- bitshift, bx <- new bitshift and cur_numbits
	jcxz	Shifted

In0:
	shr	dx, 1
	rcr	ax, 1
	loop	In0	; shift input value bitshift times

Shifted:
; now ax has the code plus a few extra bits
	mov	cx, bx	; ch <- new bitshift; cl <- cur_numbits
	sub	cl, 16
	neg	cl	; cl <- 16-cur_numbits
	shl	ah, cl
	shr	ah, cl	; throw out the top cl bits
	mov	bx, ax	; bx is the final code

; unless bitshift is 0, we've incremented one too many

	or	ch, ch	; was there an even number of bits to read?
	jz	InBDone	; yes, so input pointer is okay
	dec	si	; no point back to current byte

InBDone:
	pop	ds	; restore the registers
	pop	dx
	pop	ax

	ret
inbits	endp

;---------------------------------------------------------------------------
; bit_load_buffer
; load a buffer being used for variable length input
; The last byte in the buffer may or may not be partially
; read when this routine is called.
; RESERVE bytes are reserved at the end of the buffer. When the
; caller overruns the buffer, it calls this routine again. This code
; responds, again reserving RESERVE bytes. When the file is empty, it makes the
; RESERVE bytes available; subsequent calls will fail.
; on entry:
;	dx:si points to last addressed byte -- may be partially read
;	bx is the input file handle
; on exit:
;	dx = size of the input buffer
;	si = 0
;---------------------------------------------------------------------------
.DATA
	blb_empty	db	0	; is the file empty?
	actual_len	dw	?	; how long is it, really?

.CODE
bit_load_buffer	proc	near

	cmp	blb_empty,0	; emptied already?
	je	DoBLB
	stc	; set the carry flag and return
	ret

DoBLB:
	push	ax	; save the registers
	push	cx
	push	es

	mov	es, dx	; input buffer

	push	bx	; save the filehandle

; move actual_len-si bytes to the front of the buffer
; (if si is zero or equals actual_len, skip this step)

	xor	dx, dx	; offset to read in to 

	mov	ax, si	; see if si is zero
	or	ax, ax
	jz	SkipMove

	sub	ax, actual_len	; subtract si from actual len
	neg	ax	; leave difference in cx
	mov	cx, ax
	xor	bx, bx	; clear bx, the loop counter

	jcxz	SkipMove

BLB0:
	mov	al, es:[si]
	mov	es:[bx], al
	inc	bx
	inc	si
	loop	BLB0

	mov	dx, bx	; offset to read in to (spare the bytes we moved)

SkipMove:

	pop	bx	; filehandle
	push	ds	; save the data segment

	push	es
	pop	ds	; ds <- input buffer ptr

	mov	ah, 3Fh	; dos function number
	mov	cx, FILEBUFSIZ	; number of bytes to read
	sub	cx, dx	; fill in only what we don't have
	int	21h	; make call

	pop	ds	; restore data segment

	xor	si, si	; clear si
	add	ax, dx	; add in the reserved bytes
	cmp	ax, RESERVE	; if ax less than RESERVE, file is empty
	ja	NotEmpty
	inc	blb_empty	; set empty flag and don't reserve bytes
	jmp	BLB_IsEmpty

NotEmpty:
	mov	actual_len, ax
	sub	ax, RESERVE	; cheat bytes

BLB_IsEmpty:
	clc
	mov	dx, ax
	pop	es
	pop	cx
	pop	ax
	ret

bit_load_buffer	endp

;---------------------------------------------------------------------------
; minbits
; find the minimum number of bits to represent a word
; on entry:
;	bx - word to minbits
; on exit:
;	bx - number of bits to represent this word
;	preserve everything else
; NOTE: this code assumes return values will always be at least 9
;---------------------------------------------------------------------------
minbits	proc	near

	xor	bl, bl	; don't need to check bl, it's always set

ShiftAgain:
	inc	bl
	shr	bh, 1
	jz	MinbitsDone
	jmp	short	ShiftAgain

MinbitsDone:
	add	bl, 8
	ret

minbits	endp

;---------------------------------------------------------------------------
; dump buffer
; dump the buffer to the output file. clear the local byte
; counter and update the total byte counter.
; this routine does no disk error checking
; on entry:
;	es:0 - pointer to start of buffer
;	di - length of buffer
;	bx - output file handle
; on exit:
;    di cleared
;	global totalbytes updated
;---------------------------------------------------------------------------
dump_buffer	proc	near

	push	ax	; save all the registers
	push	cx
	push	dx

	push	ds

	mov	ah, 40h	; dos function number
	mov	cx, di	; length of buffer
	xor	dx, dx	; offset of buffer
	push	es
	pop	ds	; segment of buffer

	int	21h	; make call

	pop	ds	; data segment
	add	ax, totalbytes	; update total bytes written
	adc	totalbytes+2, 0
	mov	totalbytes, ax

	xor	di, di	; clear di

	pop	dx	; save all the registers
	pop	cx
	pop	ax

	ret
dump_buffer	endp

	END

